# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Test file handler.
"""

import shutil
import unittest
import random
import tempfile
from pathlib import Path
from datetime import datetime, timedelta

import numpy as np

from ascat.file_handling import FilenameTemplate
from ascat.file_handling import FileSearch
from ascat.file_handling import ChronFiles
from ascat.file_handling import CsvFile
from ascat.file_handling import CsvFiles


def generate_test_data():
    """
    Generate fake test data.
    """
    tmpdir = Path(tempfile.mkdtemp())

    for num in ["01", "02", "03"]:
        for tile in ["EN01212", "EN01234", "EN01256"]:
            folder = tmpdir / "temperature" / tile
            folder.mkdir(parents=True, exist_ok=True)

            filename = tmpdir / "temperature" / tile / f"202201{num}_ascat.csv"

            csv_file = CsvFile(filename, mode="w")
            file_dates = [
                datetime.strptime(f"202201{num}", "%Y%m%d") +
                timedelta(seconds=(3600 * i + seconds))
                for i, seconds in enumerate(range(3500, 3700))
            ]

            file_temps = random.choices(range(20, 35), k=15)
            dtype = np.dtype([("date", "datetime64[s]"),
                              ("temperature", "float32")])
            write_data = np.array(
                list(zip(file_dates, file_temps)), dtype=dtype)
            csv_file.write(write_data)

    for num in ["01", "02", "03"]:
        for tile in ["EN01212", "EN01234", "EN01256"]:
            prev_time = datetime.strptime(f"202201{num}0000", "%Y%m%d%H%M")
            next_day = prev_time + timedelta(days=1)

            folder = tmpdir / "precipitation" / tile
            folder.mkdir(parents=True, exist_ok=True)

            while prev_time < next_day:
                file_dates = [
                    prev_time + timedelta(seconds=i)
                    for i in range(random.randrange(120, 180))
                    if (random.randrange(10) < 7)
                ]
                for i, date in enumerate(file_dates):
                    if date >= next_day:
                        prev_time = date
                        del (file_dates[i + 1:])

                if len(file_dates) < 2:
                    continue

                d1, d2 = file_dates[0].strftime(
                    "%Y%m%d"), file_dates[-2].strftime("%Y%m%d")
                t1, t2 = file_dates[0].strftime(
                    "%H%M%S"), file_dates[-2].strftime("%H%M%S")
                filename = folder / Path(f"ascat_{d1}_{t1}-{d2}_{t2}.csv")

                csv_file = CsvFile(filename, mode="w")
                file_precips = random.choices(
                    range(0, 5), k=len(file_dates) - 1)
                dtype = np.dtype([("date", "datetime64[s]"),
                                  ("precipitation", "float32")])
                write_data = np.array(
                    list(zip(file_dates[:-1], file_precips)), dtype=dtype)
                csv_file.write(write_data)
                prev_time = file_dates[-1]

    return tmpdir


class CustomTestCase(unittest.TestCase):
    """
    Custom test case generating and deleting test data.
    """

    @classmethod
    def setUpClass(cls):
        cls.tmpdir = generate_test_data()

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmpdir)


class TestFilenameTemplate(unittest.TestCase):
    """
    Tests for FilenameTemplate class.
    """

    def setUp(self):
        """
        Setup test.
        """
        self.tmpdir = Path(tempfile.mkdtemp())
        fn_pattern = "{date}_*.{suffix}"
        sf_pattern = {"variables": "{variable}", "tiles": "{tile}"}
        self.template = FilenameTemplate(self.tmpdir, fn_pattern, sf_pattern)

    def test_template_property(self):
        """
        Test the template property.
        """
        self.assertEqual(
            str(self.template.template),
            str(self.tmpdir / "{variable}" / "{tile}" / "{date}_*.{suffix}"))

    def test_build_filename(self):
        """
        Test the build_filename method.
        """
        fn_fmt = {"date": "20220101", "suffix": "csv"}
        sf_fmt = {
            "variables": {
                "variable": "temperature"
            },
            "tiles": {
                "tile": "EN01234"
            }
        }
        self.assertEqual(
            self.template.build_filename(fn_fmt, sf_fmt),
            str(self.tmpdir / "temperature" / "EN01234" / "20220101_*.csv"))

    def test_build_basename(self):
        """
        Test the build_basename method.
        """
        fmt = {"date": "20220101", "suffix": "csv"}
        self.assertEqual(self.template.build_basename(fmt), "20220101_*.csv")

    def test_build_subfolder(self):
        """
        Test the build_subfolder method.
        """
        fmt = {
            "variables": {
                "variable": "temperature"
            },
            "tiles": {
                "tile": "EN01234"
            }
        }
        self.assertEqual(
            self.template.build_subfolder(fmt), ["temperature", "EN01234"])


class TestFileSearch(CustomTestCase):
    """
    Tests for FileSearch class.
    """

    def setUp(self):
        """
        Setup test.
        """
        fn_pattern = "{date}_*.{suffix}"
        sf_pattern = {"variables": "{variable}", "tiles": "{tile}"}
        self.filesearch = FileSearch(self.tmpdir, fn_pattern, sf_pattern)

    def test_search(self):
        """
        Test search.
        """
        fn_fmt = {"date": "20220101", "suffix": "csv"}
        sf_fmt = {
            "variables": {
                "variable": "temperature"
            },
            "tiles": {
                "tile": "EN01234"
            }
        }
        recursive = False
        search_result = self.filesearch.search(fn_fmt, sf_fmt, recursive)
        expected_result = [
            str(self.tmpdir / "temperature" / "EN01234" / "20220101_ascat.csv")
        ]
        self.assertEqual(search_result, expected_result)

    def test_search_wc(self):
        """
        Test search with wildcard.
        """
        fn_fmt = {"date": "202201*", "suffix": "csv"}
        sf_fmt = {
            "variables": {
                "variable": "temperature"
            },
            "tiles": {
                "tile": "EN012*"
            }
        }
        recursive = False
        search_result = self.filesearch.search(fn_fmt, sf_fmt, recursive)
        expected_result = [
            str(item) for item in [
                self.tmpdir / "temperature" / "EN01234" / "20220101_ascat.csv",
                self.tmpdir / "temperature" / "EN01234" / "20220102_ascat.csv",
                self.tmpdir / "temperature" / "EN01234" / "20220103_ascat.csv",
                self.tmpdir / "temperature" / "EN01212" / "20220101_ascat.csv",
                self.tmpdir / "temperature" / "EN01212" / "20220102_ascat.csv",
                self.tmpdir / "temperature" / "EN01212" / "20220103_ascat.csv",
                self.tmpdir / "temperature" / "EN01256" / "20220101_ascat.csv",
                self.tmpdir / "temperature" / "EN01256" / "20220102_ascat.csv",
                self.tmpdir / "temperature" / "EN01256" / "20220103_ascat.csv"
            ]
        ]
        self.assertEqual(sorted(search_result), sorted(expected_result))

    def test_isearch(self):
        """
        Test isearch.
        """
        fn_fmt = {"date": "20220101", "suffix": "csv"}
        sf_fmt = {
            "variables": {
                "variable": "temperature"
            },
            "tiles": {
                "tile": "EN01234"
            }
        }
        recursive = False
        search_result = self.filesearch.isearch(fn_fmt, sf_fmt, recursive)
        expected_result = iter(
            [str(self.tmpdir / "temperature" / "EN01234" / "20220101_ascat.csv")])

        self.assertEqual(list(search_result), list(expected_result))

    def test_isearch_wc(self):
        """
        Test isearch with wildcard.
        """
        fn_fmt = {"date": "202201*", "suffix": "csv"}
        sf_fmt = {
            "variables": {
                "variable": "temperature"
            },
            "tiles": {
                "tile": "EN012*"
            }
        }
        recursive = False
        search_result = self.filesearch.isearch(fn_fmt, sf_fmt, recursive)
        expected_result = iter([
            str(item) for item in [
                self.tmpdir / "temperature" /"EN01212"/ "20220101_ascat.csv",
                self.tmpdir / "temperature" /"EN01234"/ "20220101_ascat.csv",
                self.tmpdir / "temperature" /"EN01256"/ "20220101_ascat.csv",
                self.tmpdir / "temperature" /"EN01212"/ "20220102_ascat.csv",
                self.tmpdir / "temperature" /"EN01234"/ "20220102_ascat.csv",
                self.tmpdir / "temperature" /"EN01256"/ "20220102_ascat.csv",
                self.tmpdir / "temperature" /"EN01212"/ "20220103_ascat.csv",
                self.tmpdir / "temperature" /"EN01234"/ "20220103_ascat.csv",
                self.tmpdir / "temperature" /"EN01256"/ "20220103_ascat.csv"
            ]
        ])
        self.assertEqual(sorted(search_result), sorted(expected_result))

    def test_create_search_func(self):
        """
        Test custom search function.
        """

        def custom_func(arg1, arg2):
            fn_fmt = {"date": arg1, "suffix": "csv"}
            sf_fmt = {
                "variables": {
                    "variable": arg2
                },
                "tiles": {
                    "tile": "EN01234"
                }
            }
            return fn_fmt, sf_fmt

        recursive = False
        custom_search_func = self.filesearch.create_search_func(
            custom_func, recursive)
        search_result = custom_search_func("20220101", "temperature")
        expected_result = [
            str(self.tmpdir / "temperature" / "EN01234" / "20220101_ascat.csv")
        ]
        self.assertEqual(search_result, expected_result)

    def test_create_search_func_wc(self):
        """
        Test custom search function with wildcard.
        """

        def custom_func(arg1, arg2):
            fn_fmt = {"date": arg1, "suffix": "csv"}
            sf_fmt = {
                "variables": {
                    "variable": arg2
                },
                "tiles": {
                    "tile": "EN012*"
                }
            }
            return fn_fmt, sf_fmt

        recursive = False
        custom_search_func = self.filesearch.create_search_func(
            custom_func, recursive)
        search_result = custom_search_func("202201*", "temperature")
        expected_result = [
            str(item) for item in [
                self.tmpdir / "temperature" /"EN01234"/ "20220101_ascat.csv",
                self.tmpdir / "temperature" /"EN01234"/ "20220102_ascat.csv",
                self.tmpdir / "temperature" /"EN01234"/ "20220103_ascat.csv",
                self.tmpdir / "temperature" /"EN01212"/ "20220101_ascat.csv",
                self.tmpdir / "temperature" /"EN01212"/ "20220102_ascat.csv",
                self.tmpdir / "temperature" /"EN01212"/ "20220103_ascat.csv",
                self.tmpdir / "temperature" /"EN01256"/ "20220101_ascat.csv",
                self.tmpdir / "temperature" /"EN01256"/ "20220102_ascat.csv",
                self.tmpdir / "temperature" /"EN01256"/ "20220103_ascat.csv"
            ]
        ]
        self.assertEqual(sorted(search_result), sorted(expected_result))

    def test_create_isearch_func(self):
        """
        Test custom isearch function with wildcard.
        """

        def custom_func(arg1, arg2):
            fn_fmt = {"date": arg1, "suffix": "csv"}
            sf_fmt = {
                "variables": {
                    "variable": arg2
                },
                "tiles": {
                    "tile": "EN01234"
                }
            }
            return fn_fmt, sf_fmt

        recursive = False
        custom_isearch_func = self.filesearch.create_isearch_func(
            custom_func, recursive)
        search_result = custom_isearch_func("20220101", "temperature")
        expected_result = iter(
            [str(self.tmpdir / "temperature" / "EN01234" / "20220101_ascat.csv")])

        self.assertEqual(list(search_result), list(expected_result))

    def test_create_isearch_func_wc(self):
        """
        Test custom isearch function with wildcard.
        """

        def custom_func(arg1, arg2):
            fn_fmt = {"date": arg1, "suffix": "csv"}
            sf_fmt = {
                "variables": {
                    "variable": arg2
                },
                "tiles": {
                    "tile": "EN012*"
                }
            }
            return fn_fmt, sf_fmt

        recursive = False
        custom_isearch_func = self.filesearch.create_isearch_func(
            custom_func, recursive)
        search_result = custom_isearch_func("202201*", "temperature")
        expected_result = iter([
            str(item) for item in [
                self.tmpdir / "temperature" /"EN01234"/ "20220101_ascat.csv",
                self.tmpdir / "temperature" /"EN01234"/ "20220102_ascat.csv",
                self.tmpdir / "temperature" /"EN01234"/ "20220103_ascat.csv",
                self.tmpdir / "temperature" /"EN01212"/ "20220101_ascat.csv",
                self.tmpdir / "temperature" /"EN01212"/ "20220102_ascat.csv",
                self.tmpdir / "temperature" /"EN01212"/ "20220103_ascat.csv",
                self.tmpdir / "temperature" /"EN01256"/ "20220101_ascat.csv",
                self.tmpdir / "temperature" /"EN01256"/ "20220102_ascat.csv",
                self.tmpdir / "temperature" /"EN01256"/ "20220103_ascat.csv"
            ]
        ])
        self.assertEqual(sorted(search_result), sorted(expected_result))


class TestChronFiles(CustomTestCase):
    """
    Tests for ChronFiles class, as well as MultiFileHandler
    """

    def setUp(self):
        """
        Setup test.
        """
        cls = CsvFile
        fn_templ = "{date}_ascat.csv"
        sf_templ = {"variables": "{variable}", "tiles": "{tile}"}
        cls_kwargs = None
        err = True
        fn_read_fmt = lambda timestamp: {"date": timestamp.strftime("%Y%m%d")}
        sf_read_fmt = {
            "variables": {
                "variable": "temperature"
            },
            "tiles": {
                "tile": "EN01234"
            }
        }
        self.chron_files = ChronFiles(self.tmpdir, cls, fn_templ, sf_templ,
                                      cls_kwargs, err, fn_read_fmt,
                                      sf_read_fmt)

    def test_search_date(self):
        """
        Test search date.
        """
        timestamp = datetime(2022, 1, 1)
        filenames = self.chron_files.search_date(
            timestamp, date_field_fmt="%Y%m%d", date_field="date")
        expected_filenames = [
            str(self.tmpdir / "temperature" / "EN01234" / "20220101_ascat.csv")
        ]
        self.assertEqual(filenames, expected_filenames)

    def test_search_period(self):
        """
        Test search period.
        """
        dt_start = datetime(2022, 1, 1)
        dt_end = datetime(2022, 1, 3)
        filenames = self.chron_files.search_period(
            dt_start, dt_end, dt_delta=timedelta(days=1))
        expected_filenames = [
            str(self.tmpdir / "temperature" / "EN01234" / "20220101_ascat.csv"),
            str(self.tmpdir / "temperature" / "EN01234" / "20220102_ascat.csv"),
            str(self.tmpdir / "temperature" / "EN01234" / "20220103_ascat.csv")
        ]
        self.assertTrue(filenames)
        self.assertEqual(filenames, expected_filenames)

    def test_search_period_exclusive(self):
        """
        Test search period.
        """
        dt_start = datetime(2022, 1, 1)
        dt_end = datetime(2022, 1, 3)
        filenames = self.chron_files.search_period(
            dt_start, dt_end, dt_delta=timedelta(days=1), end_inclusive=False)
        expected_filenames = [
            str(self.tmpdir / "temperature" / "EN01234" / "20220101_ascat.csv"),
            str(self.tmpdir / "temperature" / "EN01234" / "20220102_ascat.csv"),
        ]
        self.assertTrue(filenames)
        self.assertEqual(filenames, expected_filenames)

    def test_read_period(self):
        """
        Test read period.
        """
        dt_start = datetime(2022, 1, 1, hour=12, minute=30)
        dt_end = datetime(2022, 1, 2, hour=12, minute=30)
        data = self.chron_files.read_period(
            dt_start, dt_end, dt_delta=timedelta(days=1))

        self.assertTrue(data["date"].min() >= dt_start)
        self.assertTrue(data["date"].max() <= dt_end)


class TestCsvFiles(unittest.TestCase):
    """
    Tests for CsvFiles class.
    """

    def setUp(self):
        """
        Setup test.
        """
        self.tmpdir = Path(tempfile.mkdtemp())
        self.csv = CsvFiles(self.tmpdir)

    def test_read_write(self):
        """
        Test read/write CSV files.
        """
        dtype = np.dtype([("date", "datetime64[h]"),
                          ("temperature", "float32")])

        dates = np.arange("2000-01-01", "2000-01-10", dtype="datetime64[D]")

        tmp_data = {}

        for date in dates:
            dt = np.arange(
                date, date + np.timedelta64(1, "D"), dtype="datetime64[h]")
            temperature = random.choices(range(0, 40), k=dt.size)
            data = np.array(list(zip(dt, temperature)), dtype=dtype)

            self.csv.write(data, date.astype(datetime))
            tmp_data[date] = data

        for date in dates:
            data = self.csv.read(date.astype(datetime))
            np.testing.assert_array_equal(data, tmp_data[date])

    def tearDown(self):
        """
        Tear down test.
        """
        shutil.rmtree(self.tmpdir)


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
Readers for ASCAT Level 1b and Level 2 data in BUFR format.
"""

import os

import numpy as np
import xarray as xr
from cadati.cal_date import cal2dt

try:
    import pdbufr
except ImportError:
    pass

from ascat.utils import tmp_unzip
from ascat.utils import mask_dtype_nans
from ascat.utils import uint8_nan
from ascat.utils import uint16_nan
from ascat.utils import int32_nan
from ascat.utils import float32_nan
from ascat.read_native import AscatFile

bufr_nan = 1.7e+38

nan_val_dict = {
    np.float32: float32_nan,
    np.uint8: uint8_nan,
    np.uint16: uint16_nan,
    np.int32: int32_nan
}

class AscatL1bBufrFile(AscatFile):
    """
    Read ASCAT Level 1b file in BUFR format.
    """

    def __init__(self, filename, **kwargs):
        """
        Initialize AscatL1bBufrFile.

        Parameters
        ----------
        filename : str
            Filename.
        """
        super().__init__(filename, **kwargs)

        for i, fname in enumerate(self.filenames):
            if os.path.splitext(fname)[1] == '.gz':
                self.filenames[i] = tmp_unzip(fname)

        self.msg_name_lookup = {
                4: "Satellite Identifier",
                6: "Direction Of Motion Of Moving Observing Platform",
                16: "Orbit Number",
                17: "Cross-Track Cell Number",
                21: "f_Beam Identifier",
                22: "f_Radar Incidence Angle",
                23: "f_Antenna Beam Azimuth",
                24: "f_Backscatter",
                25: "f_Radiometric Resolution (Noise Value)",
                26: "f_ASCAT KP Estimate Quality",
                27: "f_ASCAT Sigma-0 Usability",
                34: "f_ASCAT Land Fraction",
                35: "m_Beam Identifier",
                36: "m_Radar Incidence Angle",
                37: "m_Antenna Beam Azimuth",
                38: "m_Backscatter",
                39: "m_Radiometric Resolution (Noise Value)",
                40: "m_ASCAT KP Estimate Quality",
                41: "m_ASCAT Sigma-0 Usability",
                48: "m_ASCAT Land Fraction",
                49: "a_Beam Identifier",
                50: "a_Radar Incidence Angle",
                51: "a_Antenna Beam Azimuth",
                52: "a_Backscatter",
                53: "a_Radiometric Resolution (Noise Value)",
                54: "a_ASCAT KP Estimate Quality",
                55: "a_ASCAT Sigma-0 Usability",
                62: "a_ASCAT Land Fraction"
            }

    def _read(self, filename, generic=False, to_xarray=False):
        """
        Read one ASCAT Level 1b BUFR file.

        Parameters
        ----------
        generic : bool, optional
            'True' reading and converting into generic format or
            'False' reading original field names (default: False).
        to_xarray : bool, optional
            'True' return data as xarray.Dataset
            'False' return data as numpy.ndarray (default: False).

        Returns
        -------
        ds : xarray.Dataset, numpy.ndarray
            ASCAT Level 1b data.
        """
        df = pdbufr.read_bufr(filename, columns="data", flat=True)

        col_rename = {}
        for i, col in enumerate(df.columns.to_list()):
            name = self.msg_name_lookup.get(i + 1, None)
            if name is not None:
                col_rename[col] = name

        data = df.rename(columns=col_rename)[col_rename.values()]

        data["lat"] = df["#1#latitude"].values.astype(np.float32)
        data["lon"] = df["#1#longitude"].values.astype(np.float32)

        year = df["#1#year"].values.astype(int)
        month = df["#1#month"].values.astype(int)
        day = df["#1#day"].values.astype(int)
        hour = df["#1#hour"].values.astype(int)
        minute = df["#1#minute"].values.astype(int)
        seconds = df["#1#second"].values.astype(int)
        milliseconds = np.zeros(seconds.size)
        cal_dates = np.vstack(
            (year, month, day, hour, minute, seconds, milliseconds)).T

        data['time'] = cal2dt(cal_dates)
        data = data.to_records(index=False)
        data = {name:data[name] for name in data.dtype.names}

        metadata = {}
        metadata['platform_id'] = data['Satellite Identifier'][0].astype(int)
        metadata['orbit_start'] = np.uint32(data['Orbit Number'][0])
        metadata['filename'] = os.path.basename(filename)

        # add/rename/remove fields according to generic format
        if generic:
            data = conv_bufrl1b_generic(data, metadata)

        # convert dict to xarray.Dataset or numpy.ndarray
        if to_xarray:
            for k in data.keys():
                if len(data[k].shape) == 1:
                    dim = ['obs']
                elif len(data[k].shape) == 2:
                    dim = ['obs', 'beam']

                data[k] = (dim, data[k])

            coords = {}
            coords_fields = ['lon', 'lat', 'time']
            for cf in coords_fields:
                coords[cf] = data.pop(cf)

            data = xr.Dataset(data, coords=coords, attrs=metadata)
            if generic:
                data = mask_dtype_nans(data)
        else:
            # collect dtype info
            dtype = []
            for var_name in data.keys():
                if len(data[var_name].shape) == 1:
                    dtype.append((var_name, data[var_name].dtype.str))
                elif len(data[var_name].shape) > 1:
                    dtype.append((var_name, data[var_name].dtype.str,
                                  data[var_name].shape[1:]))

            ds = np.empty(data['time'].size, dtype=np.dtype(dtype))
            for k, v in data.items():
                ds[k] = v

            data = ds

        return data, metadata

    def _merge(self, data):
        """
        Merge data.

        Parameters
        ----------
        data : list
            List of array.

        Returns
        -------
        data : numpy.ndarray or xarray.Dataset
            Data.
        """
        if isinstance(data[0], tuple):
            data, metadata = zip(*data)
            if isinstance(data[0], xr.Dataset):
                data = xr.concat(data,
                                 dim="obs",
                                 combine_attrs="drop_conflicts")
            else:
                data = np.hstack(data)
            data = (data, metadata)
        else:
            data = np.hstack(data)
        return data

class AscatL1bBufrFileGeneric(AscatL1bBufrFile):
    """
    The same as AscatL1bBufrFile but with generic=True by default.
    """
    def _read(self, filename, generic=True, to_xarray=False, **kwargs):
        return super()._read(filename, generic=generic, to_xarray=to_xarray, **kwargs)


def conv_bufrl1b_generic(data, metadata):
    """
    Rename and convert data types of dataset.

    Spacecraft_id vs sat_id encoding

    BUFR encoding - Spacecraft_id
    - 1 ERS 1
    - 2 ERS 2
    - 3 Metop-1 (Metop-B)
    - 4 Metop-2 (Metop-A)
    - 5 Metop-3 (Metop-C)

    Internal encoding - sat_id
    - 1 ERS 1
    - 2 ERS 2
    - 3 Metop-2 (Metop-A)
    - 4 Metop-1 (Metop-B)
    - 5 Metop-3 (Metop-C)

    Parameters
    ----------
    data: dict of numpy.ndarray
        Original dataset.
    metadata: dict
        Metadata.

    Returns
    -------
    data: dict of numpy.ndarray
        Converted dataset.
    """
    skip_fields = ['Satellite Identifier']

    gen_fields_beam = {
        'Radar Incidence Angle': ('inc', np.float32, bufr_nan, 1),
        'Backscatter': ('sig', np.float32, bufr_nan, 1),
        'Antenna Beam Azimuth': ('azi', np.float32, bufr_nan, 1),
        'ASCAT Sigma-0 Usability': ('f_usable', np.uint8, None, 1),
        'Beam Identifier': ('beam_num', np.uint8, None, 1),
        'Radiometric Resolution (Noise Value)':
        ('kp', np.float32, bufr_nan, 0.01),
        'ASCAT KP Estimate Quality': ('kp_quality', np.uint8, bufr_nan, 1),
        'ASCAT Land Fraction': ('f_land', np.float32, None, 1)
    }

    gen_fields_lut = {
        'Orbit Number': ('abs_orbit_nr', np.int32),
        'Cross-Track Cell Number': ('node_num', np.uint8),
        'Direction Of Motion Of Moving Observing Platform':
        ('sat_track_azi', np.float32)
    }

    for var_name in skip_fields:
        if var_name in data:
            data.pop(var_name)

    for var_name, (new_name, new_dtype) in gen_fields_lut.items():
        data[new_name] = data.pop(var_name).astype(new_dtype)

    for var_name, (new_name, new_dtype, nan_val, s) in gen_fields_beam.items():
        f = ['{}_{}'.format(b, var_name) for b in ['f', 'm', 'a']]
        data[new_name] = np.vstack((data.pop(f[0]), data.pop(f[1]),
                                    data.pop(f[2]))).T.astype(new_dtype)
        if nan_val is not None:
            valid = data[new_name] != nan_val
            data[new_name][~valid] = nan_val_dict[new_dtype]
            data[new_name][valid] *= s

    if data['node_num'].max() == 82:
        data['swath_indicator'] = 1 * (data['node_num'] > 41)
    elif data['node_num'].max() == 42:
        data['swath_indicator'] = 1 * (data['node_num'] > 21)
    else:
        raise ValueError('Cross-track cell number size unknown')

    n_lines = data['lat'].shape[0] / data['node_num'].max()
    data['line_num'] = np.arange(n_lines).repeat(data['node_num'].max())

    sat_id = np.array([0, 0, 0, 4, 3, 5], dtype=np.uint8)
    data['sat_id'] = np.zeros(data['time'].size, dtype=np.uint8) + sat_id[int(
        metadata['platform_id'])]

    # compute ascending/descending direction
    data['as_des_pass'] = (data['sat_track_azi'] < 270).astype(np.uint8)

    return data


class AscatL2BufrFile(AscatFile):
    """
    Read ASCAT Level 2 file in BUFR format.
    """

    def __init__(self, filename, **kwargs):
        """
        Initialize AscatL2BufrFile.

        Parameters
        ----------
        filename: str
            Filename.
        """
        super().__init__(filename, **kwargs)

        for i, fname in enumerate(self.filenames):
            if os.path.splitext(fname)[1] == '.gz':
                self.filenames[i] = tmp_unzip(fname)

        self.msg_name_lookup = {
            4: "Satellite Identifier",
            6: "Direction Of Motion Of Moving Observing Platform",
            16: "Orbit Number",
            17: "Cross-Track Cell Number",
            21: "f_Beam Identifier",
            22: "f_Radar Incidence Angle",
            23: "f_Antenna Beam Azimuth",
            24: "f_Backscatter",
            25: "f_Radiometric Resolution (Noise Value)",
            26: "f_ASCAT KP Estimate Quality",
            27: "f_ASCAT Sigma-0 Usability",
            34: "f_ASCAT Land Fraction",
            35: "m_Beam Identifier",
            36: "m_Radar Incidence Angle",
            37: "m_Antenna Beam Azimuth",
            38: "m_Backscatter",
            39: "m_Radiometric Resolution (Noise Value)",
            40: "m_ASCAT KP Estimate Quality",
            41: "m_ASCAT Sigma-0 Usability",
            48: "m_ASCAT Land Fraction",
            49: "a_Beam Identifier",
            50: "a_Radar Incidence Angle",
            51: "a_Antenna Beam Azimuth",
            52: "a_Backscatter",
            53: "a_Radiometric Resolution (Noise Value)",
            54: "a_ASCAT KP Estimate Quality",
            55: "a_ASCAT Sigma-0 Usability",
            62: "a_ASCAT Land Fraction",
            65: "Surface Soil Moisture (Ms)",
            66: "Estimated Error In Surface Soil Moisture",
            67: "Backscatter",
            68: "Estimated Error In Sigma0 At 40 Deg Incidence Angle",
            69: "Slope At 40 Deg Incidence Angle",
            70: "Estimated Error In Slope At 40 Deg Incidence Angle",
            71: "Soil Moisture Sensitivity",
            72: "Dry Backscatter",
            73: "Wet Backscatter",
            74: "Mean Surface Soil Moisture",
            # 75: "Rain Fall Detection",
            76: "Soil Moisture Correction Flag",
            77: "Soil Moisture Processing Flag",
            78: "Soil Moisture Quality",
            79: "Snow Cover",
            80: "Frozen Land Surface Fraction",
            81: "Inundation And Wetland Fraction",
            82: "Topographic Complexity",
        }

    def _read(self, filename, generic=False, to_xarray=False):
        """
        Read one ASCAT Level 2 BUFR file.

        Parameters
        ----------
        generic : bool, optional
            'True' reading and converting into generic format or
            'False' reading original field names(default: False).
        to_xarray : bool, optional
            'True' return data as xarray.Dataset
            'False' return data as numpy.ndarray(default: False).

        Returns
        -------
        data : xarray.Dataset or numpy.ndarray
            ASCAT data.
        metadata : dict
            Metadata.
        """
        df = pdbufr.read_bufr(filename, columns="data", flat=True)

        col_rename = {}
        for i, col in enumerate(df.columns.to_list()):
            name = self.msg_name_lookup.get(i + 1, None)
            if name is not None:
                col_rename[col] = name

        data = df.rename(columns=col_rename)[col_rename.values()]

        data["lat"] = df["#1#latitude"].values.astype(np.float32)
        data["lon"] = df["#1#longitude"].values.astype(np.float32)

        year = df["#1#year"].values.astype(int)
        month = df["#1#month"].values.astype(int)
        day = df["#1#day"].values.astype(int)
        hour = df["#1#hour"].values.astype(int)
        minute = df["#1#minute"].values.astype(int)
        seconds = df["#1#second"].values.astype(int)
        milliseconds = np.zeros(seconds.size)
        cal_dates = np.vstack(
            (year, month, day, hour, minute, seconds, milliseconds)).T

        data['time'] = cal2dt(cal_dates)
        data = data.to_records(index=False)
        data = {name:data[name] for name in data.dtype.names}

        data["Mean Surface Soil Moisture"] *= 100.

        metadata = {}
        metadata['platform_id'] = data['Satellite Identifier'][0].astype(int)
        metadata['orbit_start'] = np.uint32(data['Orbit Number'][0])
        metadata['filename'] = os.path.basename(filename)

        # add/rename/remove fields according to generic format
        if generic:
            data = conv_bufrl2_generic(data, metadata)

        # convert dict to xarray.Dataset or numpy.ndarray
        if to_xarray:
            for k in data.keys():
                if len(data[k].shape) == 1:
                    dim = ['obs']
                elif len(data[k].shape) == 2:
                    dim = ['obs', 'beam']

                data[k] = (dim, data[k])

            coords = {}
            coords_fields = ['lon', 'lat', 'time']
            for cf in coords_fields:
                coords[cf] = data.pop(cf)

            data = xr.Dataset(data, coords=coords, attrs=metadata)
            if generic:
                data = mask_dtype_nans(data)
        else:
            # collect dtype info
            dtype = []
            # fill_value = []

            for var_name in data.keys():

                if len(data[var_name].shape) == 1:
                    dtype.append((var_name, data[var_name].dtype.str))
                    # fill_value.append(data[var_name].fill_value)

                elif len(data[var_name].shape) > 1:
                    dtype.append((var_name, data[var_name].dtype.str,
                                  data[var_name].shape[1:]))
                    # fill_value.append(data[var_name].shape[1] *
                    #                   [data[var_name].fill_value])

            ds = np.ma.empty(data['time'].size, dtype=np.dtype(dtype))
            # fill_value_arr = np.array((*fill_value, ), dtype=np.dtype(dtype))

            for k, v in data.items():
                ds[k] = v

            # ds.fill_value = fill_value_arr
            data = ds

        return data, metadata

    def _merge(self, data):
        """
        Merge data.

        Parameters
        ----------
        data : list
            List of array.

        Returns
        -------
        data : numpy.ndarray or xarray.Dataset
            Data.
        """
        if isinstance(data[0], tuple):
            data, metadata = zip(*data)
            if isinstance(data[0], xr.Dataset):
                data = xr.concat(data,
                                 dim="obs",
                                 combine_attrs="drop_conflicts")
            else:
                data = np.hstack(data)
            data = (data, metadata)
        else:
            data = np.hstack(data)
        return data

class AscatL2BufrFileGeneric(AscatL2BufrFile):
    """
    The same as AscatL1bBufrFile but with generic=True by default.
    """
    def _read(self, filename, generic=True, to_xarray=False, **kwargs):
        return super()._read(filename, generic=generic, to_xarray=to_xarray, **kwargs)

def conv_bufrl2_generic(data, metadata):
    """
    Rename and convert data types of dataset.

    Spacecraft_id vs sat_id encoding

    BUFR encoding - Spacecraft_id
    - 1 ERS 1
    - 2 ERS 2
    - 3 Metop-1 (Metop-B)
    - 4 Metop-2 (Metop-A)
    - 5 Metop-3 (Metop-C)

    Internal encoding - sat_id
    - 1 ERS 1
    - 2 ERS 2
    - 3 Metop-2 (Metop-A)
    - 4 Metop-1 (Metop-B)
    - 5 Metop-3 (Metop-C)

    Parameters
    ----------
    data: dict of numpy.ndarray
        Original dataset.
    metadata: dict
        Metadata.

    Returns
    -------
    data: dict of numpy.ndarray
        Converted dataset.
    """
    skip_fields = ['Satellite Identifier']

    gen_fields_beam = {
        'Radar Incidence Angle': ('inc', np.float32),
        'Backscatter': ('sig', np.float32),
        'Antenna Beam Azimuth': ('azi', np.float32),
        'ASCAT Sigma-0 Usability': ('f_usable', np.uint8),
        'Beam Identifier': ('beam_num', np.uint8),
        'Radiometric Resolution (Noise Value)': ('kp_noise', np.float32),
        'ASCAT KP Estimate Quality': ('kp', np.float32),
        'ASCAT Land Fraction': ('f_land', np.float32)
    }

    gen_fields_lut = {
        'Orbit Number': ('abs_orbit_nr', np.int32),
        'Cross-Track Cell Number': ('node_num', np.uint8),
        'Direction Of Motion Of Moving Observing Platform':
        ('sat_track_azi', np.float32),
        'Surface Soil Moisture (Ms)': ('sm', np.float32),
        'Estimated Error In Surface Soil Moisture': ('sm_noise', np.float32),
        'Backscatter': ('sig40', np.float32),
        'Estimated Error In Sigma0 At 40 Deg Incidence Angle':
        ('sig40_noise', np.float32),
        'Slope At 40 Deg Incidence Angle': ('slope40', np.float32),
        'Estimated Error In Slope At 40 Deg Incidence Angle':
        ('slope40_noise', np.float32),
        'Soil Moisture Sensitivity': ('sm_sens', np.float32),
        'Dry Backscatter': ('dry_sig40', np.float32),
        'Wet Backscatter': ('wet_sig40', np.float32),
        'Mean Surface Soil Moisture': ('sm_mean', np.float32),
        # 'Rain Fall Detection': ('rf', np.float32),
        'Soil Moisture Correction Flag': ('corr_flag', np.uint8),
        'Soil Moisture Processing Flag': ('proc_flag', np.uint8),
        'Soil Moisture Quality': ('agg_flag', np.uint8),
        'Snow Cover': ('snow_prob', np.uint8),
        'Frozen Land Surface Fraction': ('frozen_prob', np.uint8),
        'Inundation And Wetland Fraction': ('wetland', np.uint8),
        'Topographic Complexity': ('topo', np.uint8)
    }

    for var_name in skip_fields:
        if var_name in data:
            data.pop(var_name)

    for var_name, (new_name, new_dtype) in gen_fields_lut.items():
        mask = (data[var_name] == bufr_nan) |  (np.isnan(data[var_name]))
        data[var_name][mask] = nan_val_dict[new_dtype]

        data[new_name] = np.ma.array(data.pop(var_name).astype(new_dtype), mask=mask)
        data[new_name].fill_value = nan_val_dict[new_dtype]

    for var_name, (new_name, new_dtype) in gen_fields_beam.items():

        f = ['{}_{}'.format(b, var_name) for b in ['f', 'm', 'a']]

        mask = np.vstack((data[f[0]] == bufr_nan, data[f[1]] == bufr_nan,
                          data[f[2]] == bufr_nan)).T

        data[new_name] = np.ma.vstack((data.pop(f[0]), data.pop(f[1]),
                                       data.pop(f[2]))).T.astype(new_dtype)

        data[new_name].mask = mask
        data[new_name][mask] = nan_val_dict[new_dtype]

        data[new_name].fill_value = nan_val_dict[new_dtype]

    if data['node_num'].max() == 82:
        data['swath_indicator'] = np.ma.array(1 * (data['node_num'] > 41),
                                              dtype=np.uint8,
                                              mask=data['node_num'] > 82)
    elif data['node_num'].max() == 42:
        data['swath_indicator'] = np.ma.array(1 * (data['node_num'] > 21),
                                              dtype=np.uint8,
                                              mask=data['node_num'] > 42)
    else:
        raise ValueError('Cross-track cell number size unknown')

    n_lines = data['lat'].shape[0] / data['node_num'].max()
    line_num = np.arange(n_lines).repeat(data['node_num'].max())
    data['line_num'] = np.ma.array(line_num,
                                   dtype=np.uint16,
                                   mask=np.zeros_like(line_num),
                                   fill_value=uint16_nan)

    sat_id = np.ma.array([0, 0, 0, 4, 3, 5], dtype=np.uint8)
    data['sat_id'] = np.ma.zeros(data['time'].size,
                                 dtype=np.uint8) + sat_id[int(
                                     metadata['platform_id'])]
    data['sat_id'].mask = np.zeros(data['time'].size)
    data['sat_id'].fill_value = uint8_nan

    # compute ascending/descending direction
    data['as_des_pass'] = np.ma.array(data['sat_track_azi'] < 270,
                                      dtype=np.uint8,
                                      mask=np.zeros(data['time'].size),
                                      fill_value=uint8_nan)

    mask = data['lat'] == bufr_nan
    data['lat'] = np.ma.array(data['lat'], mask=mask, fill_value=float32_nan)

    mask = data['lon'] == bufr_nan
    data['lon'] = np.ma.array(data['lon'], mask=mask, fill_value=float32_nan)

    data['time'] = np.ma.array(data['time'], mask=mask, fill_value=0)

    return data



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
import argparse
from pathlib import Path

import numpy as np
import xarray as xr
from pyresample import kd_tree, SwathDefinition

from ascat.grids.grid_registry import GridRegistry
from ascat.product_info import get_swath_product_id
from ascat.product_info import swath_io_catalog
from ascat.regrid.regrid import retrieve_or_store_grid_lut


def parse_args_swath_resample(args):
    """
    Parse command line arguments for resampling an ASCAT swath file
    to a regular grid.

    Parameters
    ----------
    args : list
        Command line arguments.

    Returns
    -------
    parser : ArgumentParser
        Argument Parser object.
    """
    parser = argparse.ArgumentParser(
        description="Resample an ASCAT swath file to a regular grid")
    parser.add_argument(
        "filepath", metavar="FILEPATH", help="Path to file or folder")
    parser.add_argument(
        "outpath", metavar="OUTPATH", help="Path to the output data")
    parser.add_argument(
        "resample_deg",
        metavar="RESAMPLE_DEG",
        type=float,
        help="Target grid spacing in degrees")
    parser.add_argument(
        "--product_id",
        metavar="PRODUCT_ID",
        help="Product identifier (e.g. H129, H125, H121, etc.). If not provided, an attempt is made to determine it from the file name.")
    parser.add_argument(
        "--grid_store",
        metavar="GRID_STORE",
        help="Path for storing/loading lookup tables")
    parser.add_argument(
        "--suffix",
        metavar="SUFFIX",
        help="File suffix (default: _RESAMPLE_DEGdeg)")
    parser.add_argument(
        "--neighbours",
        metavar="N",
        type=int,
        default=6,
        help="Number of neighbours to consider for each grid point (default: 6)"
    )
    parser.add_argument(
        "--radius",
        metavar="RADIUS",
        type=float,
        default=10000,
        help="Cut off distance in meters (default: 10000)")

    return parser.parse_args(args)


def swath_resample_main(cli_args):
    """
    Read command line arguments and perform an inverse distance resampling
    procedure on a single ASCAT swath file or directory of swath files.

    Parameters
    ----------
    cli_args : list
        Command line arguments.
    """
    args = parse_args_swath_resample(cli_args)

    product_id = args.product_id

    filepath = Path(args.filepath)
    trg_grid_size = args.resample_deg

    k = args.neighbours
    radius = args.radius

    outpath = Path(args.outpath)
    outpath.parent.mkdir(parents=True, exist_ok=True)

    if args.suffix:
        suffix = args.suffix
    else:
        suffix = f"_{args.resample_deg}deg"

    if args.grid_store:
        grid_store = Path(args.grid_store)
        grid_store.parent.mkdir(parents=True, exist_ok=True)
    else:
        grid_store = None

    inverse_distance_resampling(filepath, outpath, trg_grid_size, suffix, k,
                                radius, grid_store, product_id=product_id)


def inverse_distance_resampling(filepath,
                                outpath,
                                sampling,
                                suffix,
                                k=6,
                                radius=10000.,
                                grid_store=None,
                                product_id=None):
    """
    Inverse distance resampling of ASCAT swath data.

    Parameters
    ----------
    filepath : pathlib.Path
        Path to file or folder.
    outpath : pathlib.Path
        Path to the output data.
    sampling : float
        Target grid spacing in degrees.
    suffix : str
        Filename suffix.
    k : int, optional
        Number of neighbours to consider for each grid point (default: 6)
    radius : float, optional
        Cut off distance in meters (default: 10000.)
    grid_store : pathlib.Path, optional
        Path for storing/loading lookup tables (default: None).
    product_id : str, optional
        Product identifier (e.g. H129, H125, H121, etc.). If not provided,
        an attempt is made to determine it from the file name.
    """
    if filepath.is_dir():
        files = list(filepath.glob("**/*.nc"))
    elif filepath.is_file() and filepath.suffix == ".nc":
        files = [filepath]
    else:
        raise RuntimeError("No files found at the provided filepath")

    first_file = files[0]
    product_id = product_id or get_swath_product_id(str(first_file.name))

    registry = GridRegistry()
    product = swath_io_catalog[product_id]
    src_grid = registry.get(product.grid_name)
    src_grid_size = src_grid.res
    src_grid_id = f"fib_grid_{src_grid_size}km"

    if (radius < 1000) or (radius > 100000):
        raise ValueError(f"Radius outside limits: 1000 < {radius} < 100000")

    if (k < 1) or (k > 100):
        raise ValueError(
            f"Number of neighbours outside limits 1 < {k} < {100}")

    trg_grid_id = f"reg_grid_{sampling}deg"

    trg_grid, grid_lut = retrieve_or_store_grid_lut(src_grid, src_grid_id,
                                                    trg_grid_id, sampling,
                                                    grid_store)

    var_list = [
        ("time", "nn"),
        ("as_des_pass", "nn"),
        ("swath_indicator", "nn"),
        ("surface_flag", "nn"),
        ("surface_flag_source", "nn"),
        ("surface_soil_moisture", "idw"),
        ("surface_soil_moisture_noise", "idw"),
        ("surface_soil_moisture_sensitivity", "idw"),
        ("backscatter40", "idw"),
        ("slope40", "idw"),
        ("curvature40", "idw"),
        ("snow_cover_probability", "idw"),
        ("frozen_soil_probability", "idw"),
        ("topographic_complexity", "idw"),
        ("wetland_fraction", "idw"),
        ("subsurface_scattering_prob", "idw"),
        # ("processing_flag", "major"),
        ("correction_flag", "bitwise_or"),
        ("backscatter40_flag", "bitwise_or"),
    ]

    target_def = SwathDefinition(lons=trg_grid.arrlon, lats=trg_grid.arrlat)
    output_shape = trg_grid.shape
    dim = ("latitude", "longitude")

    for f in files:
        outfile = outpath / Path(f.stem + suffix + f.suffix)

        with xr.open_dataset(f, decode_cf=False, mask_and_scale=False) as ds:

            lons = ds["longitude"] * ds["longitude"].scale_factor
            lats = ds["latitude"] * ds["latitude"].scale_factor

            coords = {
                "latitude": np.int32(trg_grid.lat2d[:, 0] / 1e-6),
                "longitude": np.int32(trg_grid.lon2d[0] / 1e-6)
            }

            resampled_ds = xr.Dataset(coords=coords)
            resampled_ds.attrs = ds.attrs

            resampled_ds["latitude"].attrs = ds["latitude"].attrs
            resampled_ds["longitude"].attrs = ds["longitude"].attrs

            swath_def = SwathDefinition(lons=lons, lats=lats)
            valid_input_index, valid_output_index, index_array, distance_array = \
                    kd_tree.get_neighbour_info(swath_def, target_def,
                                                radius, neighbours=k)

            invalid_pos = index_array == lons.size
            index_array = index_array.astype(np.int32)
            index_array[invalid_pos] = -1

            for var, method in var_list:

                if var not in ds or len(ds[var].dims) == 0:
                    continue

                data = ds[var].data[valid_input_index][index_array]
                data[invalid_pos] = ds[var]._FillValue

                resam_data = np.zeros(
                    data.shape[0], dtype=ds[var].dtype) + ds[var]._FillValue

                if method == "idw":
                    p = 2
                    weights = 1 / (distance_array**p)
                    invalid = invalid_pos | (data == ds[var]._FillValue)
                    weights[invalid] = 0
                    total_weights = weights.sum(axis=1)
                    idx = total_weights != 0
                    resam_data[idx] = np.sum(
                        weights * data, axis=1)[idx] / total_weights[idx]
                elif method == "nn":
                    valid = index_array != -1
                    first_idx = np.where(
                        valid.any(axis=1), valid.argmax(axis=1), -1)
                    valid_mask = first_idx != -1
                    valid_rows = np.arange(index_array.shape[0])[valid_mask]
                    resam_data[valid_rows] = data[valid_rows,
                                                  first_idx[valid_rows]]
                elif method == "bitwise_or":
                    mask = index_array == -1
                    data[mask] = 0
                    resam_data = np.bitwise_or.reduce(data, axis=1)
                else:
                    raise ValueError("Resampling method unknown")

                resampled_ds[var] = (dim, resam_data.reshape(output_shape))
                resampled_ds[var].attrs = ds[var].attrs
                resampled_ds[var].encoding = {"zlib": True, "complevel": 4}

            resampled_ds.to_netcdf(outfile)


def run_swath_resample():
    """Run command line interface for reample ASCAT swath data."""
    swath_resample_main(sys.argv[1:])



#!/usr/bin/env python3

import unittest
from pathlib import Path
from datetime import datetime
from tempfile import TemporaryDirectory
from time import time

import xarray as xr
import numpy as np

import ascat.read_native.generate_test_data as gtd

from ascat.swath import Swath
from ascat.swath import SwathGridFiles
from ascat.product_info import AscatH129Swath
from get_path import get_testdata_path

TESTDATA_PATH = get_testdata_path()

def gen_dummy_swathfiles(directory, sat_name=None):
    if sat_name is not None:
        directory = directory / sat_name
        directory.mkdir(parents=True, exist_ok=True)
    gtd.swath_ds.to_netcdf(directory / "swath.nc")
    gtd.swath_ds_2.to_netcdf(directory / "swath_2.nc")


class TestSwath(unittest.TestCase):
    def setUp(self):
        self.tempdir = TemporaryDirectory()
        self.tempdir_path = Path(self.tempdir.name)
        self.real_swaths_path = Path(
            TESTDATA_PATH / "hsaf/h129/swaths/metop_a/2021/01"
        )
        gen_dummy_swathfiles(self.tempdir_path)

    def tearDown(self):
        self.tempdir.cleanup()

    def test_init(self):
        swath_path = self.tempdir_path / "swath.nc"
        ra = Swath(swath_path)
        self.assertEqual(ra.filenames[0], swath_path)

        # ra_chunked = SwathFile(swath_path, chunks={"locations": 2})
        # self.assertEqual(ra_chunked.filename, swath_path)
        # self.assertIsNone(ra_chunked.ds)
        # self.assertEqual(ra_chunked.chunks, {"locations": 2})

    def test_read(self):
        # Basic open of dummy file
        swath_path = self.tempdir_path / "swath.nc"
        ra = Swath(swath_path)
        ds = ra.read()
        self.assertIsInstance(ds, xr.Dataset)
        self.assertIn("longitude", ds)
        self.assertIn("latitude", ds)
        # self.assertIn("time", ra.ds)
        self.assertIn("obs", ds.dims)

        # Open of multiple real files
        swath_paths = list(self.real_swaths_path.glob("*.nc"))
        ra = Swath(swath_paths)
        ds1 = ra.read().load()
        ds2 = ra.read(parallel=True).load()

        reconstructed = xr.open_mfdataset(
            swath_paths,
            concat_dim="obs",
            combine="nested",
            combine_attrs="drop_conflicts",
        ).load()

        ds3 = [data for data in ra.iter_read_nbytes(1_000_000_000_000)][0].load()

        xr.testing.assert_identical(ds1, reconstructed)
        xr.testing.assert_identical(ds2, reconstructed)
        xr.testing.assert_identical(ds3, reconstructed)

    def test__ensure_obs(self):
        swath_path = self.tempdir_path / "swath.nc"
        ra = Swath(swath_path)
        ds = xr.open_dataset(swath_path)
        ds = ra._ensure_obs(ds)
        self.assertIn("obs", ds.dims)
        self.assertNotIn("time", ds.dims)
        # print(original_dim.values)
        # np.testing.assert_equal(ds["obs"].values, original_dim.values)

    def test_merge(self):
        fname1 = self.tempdir_path / "swath.nc"
        fname2 = self.tempdir_path / "swath_2.nc"
        ra1 = Swath(fname1)
        ra2 = Swath(fname2)
        ds1 = ra1.read()
        ds2 = ra2.read()
        merged = ra1.merge([ds1, ds2])
        self.assertTrue(
            np.all(
                merged["location_id"].values
                == np.concatenate(
                    [ds1["location_id"].values, ds2["location_id"].values]
                )
            )
        )

    def test__nbytes(self):
        swath_path = self.tempdir_path / "swath.nc"
        ra = Swath(swath_path)
        ds = ra.read()
        nbytes = ra._nbytes(ds)
        self.assertGreater(nbytes, 0)

    def test_iter_read_nbytes(self):
        # TODO adapt and move to test_file_handling.TestFilenames

        fname1 = self.tempdir_path / "swath.nc"
        fname2 = self.tempdir_path / "swath_2.nc"
        ra = Swath([fname1, fname2])

        iterations = list(ra.iter_read_nbytes(max_nbytes=0))
        assert len(iterations) == 2

        ds1, ds2 = iterations
        total_size = ra._nbytes(ds1) + ra._nbytes(ds2)

        iterations = list(ra.iter_read_nbytes(max_nbytes=total_size))
        assert len(iterations) == 1

        iterations = list(ra.iter_read_nbytes(max_nbytes=total_size - 1))
        assert len(iterations) == 2


class TestSwathGridFiles(unittest.TestCase):
    def setUp(self):
        self.tempdir = TemporaryDirectory()
        self.tempdir_path = Path(self.tempdir.name)
        # gen_dummy_swathfiles(self.tempdir_path)

    def tearDown(self):
        self.tempdir.cleanup()

    def test_init(self):

        def _fn_read_fmt(timestamp, sat="[ABC]"):
            sat = sat.upper()
            return {
                "date": timestamp.strftime("%Y%m%d*"),
                "sat": sat,
                "placeholder": "*",
                "placeholder1": "*",
            }

        def _sf_read_fmt(timestamp, sat="[abc]"):
            sat = sat.lower()
            output_fmt = {
                "satellite_folder": {"satellite": f"metop_{sat}"},
                "year_folder": {"year": f"{timestamp.year}"},
                "month_folder": {"month": f"{timestamp.month:02d}"},
            }
            # if sat is not None:
            #     output_fmt["sat_folder"] = {
            #         "sat": f"metop_{sat.lower()}"
            #     }

            return output_fmt

        # we can create a SwathGridFiles object that points directly to a directory
        # and read the files within it, without passing a "sat" argument to
        # sf.search_period().
        swath_path = TESTDATA_PATH / "hsaf/h129/swaths/metop_a/2021/01"

        sf = SwathGridFiles(
            swath_path,
            fn_templ="W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP{sat}-6.25km-H129_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc",
            sf_templ={"satellite_folder": "metop_[abc]", "year_folder": "{year}"},
            date_field_fmt="%Y%m%d%H%M%S",
            grid_name="fibgrid_6.25",
            fn_read_fmt=_fn_read_fmt,
            sf_read_fmt=_sf_read_fmt,
        )

        files = sf.search_period(
            datetime(2021, 1, 15),
            datetime(2021, 1, 30),
            date_field_fmt="%Y%m%d%H%M%S",
            # sat="a"
        )
        self.assertGreater(len(files), 0)

        # We can also set the path to the root of the product directory and
        # optionally specify "sat" to search_period() to filter the files by
        # satellite.
        # This would be a regex, so if we just wanted metop b and c we could
        # pass sat="[bc]".
        # The default value is "[abc]" which will take all three (or whatever
        # is available)
        swath_path = TESTDATA_PATH / "hsaf/h129/swaths"
        sf = SwathGridFiles(
            swath_path,
            fn_templ="W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP{sat}-6.25km-H129_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc",
            sf_templ={"satellite_folder": "metop_[abc]", "year_folder": "{year}", "month_folder": "{month}"},
            date_field_fmt="%Y%m%d%H%M%S",
            grid_name="fibgrid_6.25",
            fn_read_fmt=_fn_read_fmt,
            sf_read_fmt=_sf_read_fmt,
        )

        files = sf.search_period(
            datetime(2021, 1, 15),
            datetime(2021, 1, 30),
            date_field_fmt="%Y%m%d%H%M%S",
            sat="a",
        )
        self.assertGreater(len(files), 0)

    def test_from_product_id(self):
        swath_path = TESTDATA_PATH / "hsaf/h129/swaths"
        sf = SwathGridFiles.from_product_id(swath_path, "h129")
        files = sf.search_period(
            datetime(2021, 1, 15), datetime(2021, 1, 30), date_field_fmt="%Y%m%d%H%M%S"
        )
        print(files)
        self.assertGreater(len(files), 0)

    def test_from_io_class(self):
        swath_path = TESTDATA_PATH / "hsaf/h129/swaths"
        sf = SwathGridFiles.from_product_class(swath_path, AscatH129Swath)
        files = sf.search_period(
            datetime(2021, 1, 15), datetime(2021, 1, 30), date_field_fmt="%Y%m%d%H%M%S"
        )
        self.assertGreater(len(files), 0)

    def test_read(self):
        # test read
        swath_path = TESTDATA_PATH / "hsaf/h129/swaths"
        sf = SwathGridFiles.from_product_id(swath_path, "h129")
        files = sf.search_period(
            datetime(2021, 1, 15), datetime(2021, 1, 30), date_field_fmt="%Y%m%d%H%M%S"
        )
        self.assertGreater(len(files), 0)
        bbox = (-180, -4, -70, 20)

        merged_ds = sf.read(
            (datetime(2021, 1, 15), datetime(2021, 1, 30)),
            bbox=bbox,
        )
        merged_ds.load()

        self.assertLess(merged_ds.time.max().values, np.datetime64(datetime(2021, 1, 30)))
        self.assertGreater(merged_ds.time.min().values, np.datetime64(datetime(2021, 1, 15)))
        self.assertLess(merged_ds.latitude.max().values, bbox[1])
        self.assertGreater(merged_ds.latitude.min().values, bbox[0])
        self.assertLess(merged_ds.longitude.max().values, bbox[3])
        self.assertGreater(merged_ds.longitude.min().values, bbox[2])

        # test extract from main folder
        swath_path = TESTDATA_PATH / "hsaf/h129/swaths/metop_a/2021/01"
        sf = SwathGridFiles.from_product_id(swath_path, "h129")
        files = sf.search_period(
            datetime(2021, 1, 15), datetime(2021, 1, 30), date_field_fmt="%Y%m%d%H%M%S"
        )
        self.assertGreater(len(files), 0)
        bbox = (-180, -4, -70, 20)

        merged_ds = sf.read(
            (datetime(2021, 1, 15), datetime(2021, 1, 30)),
            bbox=bbox,
        )
        merged_ds.load()

        self.assertLess(merged_ds.time.max().values, np.datetime64(datetime(2021, 1, 30)))
        self.assertGreater(merged_ds.time.min().values, np.datetime64(datetime(2021, 1, 15)))
        self.assertLess(merged_ds.latitude.max().values, bbox[1])
        self.assertGreater(merged_ds.latitude.min().values, bbox[0])
        self.assertLess(merged_ds.longitude.max().values, bbox[3])
        self.assertGreater(merged_ds.longitude.min().values, bbox[2])


    # def test__trim_to_gpis(self):
    #     swath_path = self.tempdir_path / "swath.nc"
    #     ra = Swath(swath_path)
    #     ds = xr.open_dataset(swath_path)
    #     ds = ds.chunk({"obs": 1_000_000})
    #     valid_gpis = [1100178, 1102762]
    #     grid = FibGrid(12.5)

    #     gpi_lookup = np.zeros(grid.gpis.max()+1, dtype=bool)
    #     gpi_lookup[valid_gpis] = 1
    #     # trimmed_ds = ra._trim_to_gpis(ds, valid_gpis)
    #     trimmed_ds = ra._trim_to_gpis(ds, lookup_vector=gpi_lookup)
    #     self.assertTrue(np.all(np.isin(trimmed_ds["location_id"].values, valid_gpis)))
    #     new_obs_gpis = trimmed_ds["location_id"].values
    #     np.testing.assert_array_equal(new_obs_gpis, np.repeat(np.array(valid_gpis), [1, 1]))

    # def test__trim_var_range(self):
    #     swath_path = self.tempdir_path / "swath.nc"
    #     ra = Swath(swath_path)
    #     ds = xr.open_dataset(swath_path)
    #     ds = ra._ensure_obs(ds)
    #     ds = ds.chunk({"obs": 1_000_000})
    #     start_dt = np.datetime64("2020-11-15T09:04:50")
    #     end_dt = np.datetime64("2020-11-15T09:04:52")
    #     trimmed_ds = ra._trim_var_range(ds, "time", start_dt, end_dt)
    #     self.assertTrue(np.all(trimmed_ds["time"].values < end_dt))
    #     self.assertTrue(np.all(trimmed_ds["time"].values >= start_dt))

    def test_stack_to_cell_files(self):
        swath_path = TESTDATA_PATH / "hsaf/h129/swaths/"
        sf = SwathGridFiles.from_product_id(swath_path, "h129")
        out_dir = self.tempdir_path / "cells_out"
        out_dir.mkdir(parents=True, exist_ok=True)
        cells_to_test = [2587, 2588]
        sf.stack_to_cell_files(
            out_dir,
            4 * (1024**3),
            date_range=(datetime(2021, 1, 1), datetime(2021, 1, 15)),
            cells=cells_to_test,
        )

        # assert that the cell files were created ( and no others )
        assert len(list(out_dir.rglob("*.nc"))) == len(cells_to_test)
        assert all(
            [
                f"{c}.nc" in [f.name for f in out_dir.rglob("*.nc")]
                for c in cells_to_test
            ]
        )

        # assert that they are not empty
        for cell, cell_file in zip(cells_to_test, out_dir.rglob("*.nc")):
            ds = xr.open_dataset(cell_file, decode_cf=True, mask_and_scale=False)
            assert len(ds.obs) > 0
            # assert that they contain all the data that the swaths did (for these cells)
            cell_ds_from_swath = sf.read(
                date_range=(datetime(2021, 1, 1), datetime(2021, 1, 15)),
                cell=[cell],
                read_kwargs={"mask_and_scale": False, "decode_cf": True},
            )
            assert len(cell_ds_from_swath.obs) == len(ds.obs)
            for variable in cell_ds_from_swath.data_vars:
                if variable in ["location_id", "lat", "lon", "latitude", "longitude"]:
                    continue
                np.testing.assert_array_equal(
                    cell_ds_from_swath[variable].values, ds[variable].values
                )
                xr.testing.assert_identical(
                    cell_ds_from_swath[variable].reset_coords(drop=True),
                    ds[variable].reset_coords(drop=True),
                )

        # reprocess to contiguous
        from ascat.cell import CellGridFiles
        cf = CellGridFiles.from_product_id(out_dir, "h129")

        contig_out_dir = self.tempdir_path / "contig_cells_out"
        contig_out_dir.mkdir(parents=True, exist_ok=True)

        cf.convert_to_contiguous(contig_out_dir, parallel=True)

        # assert that the cell files were created ( and no others )
        assert len(list(contig_out_dir.rglob("*.nc"))) == len(cells_to_test)
        assert all(
            [
                f"{c}.nc" in [f.name for f in contig_out_dir.rglob("*.nc")]
                for c in cells_to_test
            ]
        )

        # assert that the data is the same as the original
        for cell, cell_file in zip(cells_to_test, out_dir.rglob("*.nc")):
            idx_ds = xr.open_dataset(cell_file, decode_cf=True, mask_and_scale=False).cf_geom.to_indexed_ragged()
            ctg_ds = xr.open_dataset(contig_out_dir / f"{cell}.nc", decode_cf=True, mask_and_scale=False,)
            round_trip_ds = ctg_ds.cf_geom.to_indexed_ragged()

            # time will not be sorted after a round trip
            idx_ds = idx_ds.sortby("time")
            round_trip_ds = round_trip_ds.sortby("time")

            for var in idx_ds.variables:
                xr.testing.assert_identical(idx_ds[var], round_trip_ds[var])



# Copyright (c) 2025, TU Wien
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of TU Wien, Department of Geodesy and Geoinformation
#      nor the names of its contributors may be used to endorse or promote
#      products derived from this software without specific prior written
#      permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL TU WIEN DEPARTMENT OF GEODESY AND
# GEOINFORMATION BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import re

import numpy as np

from ascat.cell import RaggedArrayTs
from ascat.cell import OrthoMultiTimeseriesCell


class BaseCellProduct:
    fn_format = "{:04d}.nc"

    @classmethod
    def preprocessor(cls, ds):
        return ds


class RaggedArrayCellProduct(BaseCellProduct):
    file_class = RaggedArrayTs
    sample_dim = "obs"
    instance_dim = "locations"

    @classmethod
    def preprocessor(cls, ds):
        if "row_size" in ds.variables:
            ds["row_size"].attrs["sample_dimension"] = cls.sample_dim
        if "locationIndex" in ds.variables:
            ds["locationIndex"].attrs["instance_dimension"] = cls.instance_dim
        if "location_id" in ds.variables:
            ds["location_id"].attrs["cf_role"] = "timeseries_id"
        if ds.attrs.get("featureType") is None:
            ds = ds.assign_attrs({"featureType": "timeSeries"})
        if ds.attrs.get("grid_mapping_name") is None:
            ds.attrs["grid_mapping_name"] = cls.grid_name
        return ds


class ErsCell(RaggedArrayCellProduct):

    @classmethod
    def preprocessor(cls, ds):
        if "obs" in ds.dims:
            chunk_dim = "obs"
        else:
            chunk_dim = "time"
        ds = super().preprocessor(ds).chunk({chunk_dim: -1})
        for var in ds.variables:
            if ds[var].dtype == np.float32:
                ds[var] = ds[var].where(ds[var] > -2147483600)
            if var == "alt":
                ds[var] = ds[var].where(ds[var] < 999999)

            parts = var.split("_")
            if parts[0] in ["fore", "mid", "aft"]:
                if parts[0] == "fore":
                    parts[0] = "for"
                ds = ds.rename({var: "_".join(parts[1:] + [parts[0]])})
        return ds


class ErsHCell(ErsCell):
    grid_name = "fibgrid_12.5"


class ErsNCell(ErsCell):
    grid_name = "fibgrid_25"


class AscatH129Cell(RaggedArrayCellProduct):
    grid_name = "fibgrid_6.25"


class AscatH130Cell(RaggedArrayCellProduct):
    grid_name = "fibgrid_6.25"


class AscatH122Cell(RaggedArrayCellProduct):
    grid_name = "fibgrid_6.25"


class AscatH121Cell(RaggedArrayCellProduct):
    grid_name = "fibgrid_12.5"


class AscatH139Cell(RaggedArrayCellProduct):
    grid_name = "fibgrid_12.5"


class AscatH29Cell(RaggedArrayCellProduct):
    grid_name = "fibgrid_12.5"


class AscatSIG0Cell6250m(RaggedArrayCellProduct):
    grid_name = "fibgrid_6.25"


class AscatSIG0Cell12500m(RaggedArrayCellProduct):
    grid_name = "fibgrid_12.5"


class OrthoMultiArrayCellProduct(BaseCellProduct):
    file_class = OrthoMultiTimeseriesCell
    sample_dim = "obs"
    instance_dim = "locations"

    @classmethod
    def preprocessor(cls, ds):
        if "location_id" in ds.variables:
            ds["location_id"].attrs["cf_role"] = "timeseries_id"
        if ds.attrs.get("featureType") is None:
            ds = ds.assign_attrs({"featureType": "timeSeries"})
        return ds


class SwathProduct:
    from ascat.swath import Swath
    file_class = Swath


class AscatSwathProduct(SwathProduct):
    grid_name = None

    @classmethod
    def preprocess_(cls, ds):
        ds["location_id"].attrs["cf_role"] = "timeseries_id"
        ds.attrs["global_attributes_flag"] = 1
        ds.attrs["featureType"] = "point"
        ds.attrs["grid_mapping_name"] = cls.grid_name
        if "spacecraft" in ds.attrs:
            # Assumption: the spacecraft attribute is something like "metop-a"
            sat_id = {"a": 3, "b": 4, "c": 5}
            sat = ds.attrs["spacecraft"][-1].lower()
            ds["sat_id"] = ("obs",
                            np.repeat(np.int8(sat_id[sat]), ds["location_id"].size))
            del ds.attrs["spacecraft"]
        return ds

    @staticmethod
    def postprocess_(ds):
        for key, item in {
                "latitude": "lat",
                "longitude": "lon",
                "altitude": "alt"
        }.items():
            if key in ds:
                ds = ds.rename({key: item})
        if "altitude" not in ds:
            ds["alt"] = ("locations",
                         np.full_like(ds["lat"], fill_value=np.nan))
        return ds


class AscatH129Swath(AscatSwathProduct):
    fn_pattern = "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP{sat}-6.25km-H129_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc"
    sf_pattern = {
        "satellite_folder": "metop_[abc]",
        "year_folder": "{year}",
        "month_folder": "{month}"
    }
    date_field_fmt = "%Y%m%d%H%M%S"
    grid_name = "fibgrid_6.25"
    cell_fn_format = "{:04d}.nc"

    @staticmethod
    def fn_read_fmt(timestamp, sat="[ABC]"):
        sat = sat.upper()
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": sat,
            "placeholder": "*",
            "placeholder1": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp, sat="[abc]"):
        sat = sat.lower()
        return {
            "satellite_folder": {
                "satellite": f"metop_{sat}"
            },
            "year_folder": {
                "year": f"{timestamp.year}"
            },
            "month_folder": {
                "month": f"{timestamp.month}".zfill(2)
            },
        }


class AscatH121Swath(AscatSwathProduct):
    fn_pattern = "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP{sat}-12.5km-H121_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc"
    sf_pattern = {
        "satellite_folder": "metop_[abc]",
        "year_folder": "{year}",
        "month_folder": "{month}"
    }
    date_field_fmt = "%Y%m%d%H%M%S"
    grid_name = "fibgrid_12.5"
    cell_fn_format = "{:04d}.nc"

    @staticmethod
    def fn_read_fmt(timestamp, sat="[ABC]"):
        sat = sat.upper()
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": sat,
            "placeholder": "*",
            "placeholder1": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp, sat="[abc]"):
        sat = sat.lower()
        return {
            "satellite_folder": {
                "satellite": f"metop_{sat}"
            },
            "year_folder": {
                "year": f"{timestamp.year}"
            },
            "month_folder": {
                "month": f"{timestamp.month}".zfill(2)
            },
        }


class AscatH122Swath(AscatSwathProduct):
    fn_pattern = "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP{sat}-6.25km-H122_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc"
    sf_pattern = {}
    date_field_fmt = "%Y%m%d%H%M%S"
    grid_name = "fibgrid_6.25"
    cell_fn_format = "{:04d}.nc"

    @staticmethod
    def fn_read_fmt(timestamp, sat="[ABC]"):
        sat = sat.upper()
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": sat,
            "placeholder": "*",
            "placeholder1": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp):
        return {}


class AscatH29Swath(AscatSwathProduct):
    fn_pattern = "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP{sat}-12.5km-H29_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc"
    sf_pattern = {}
    date_field_fmt = "%Y%m%d%H%M%S"
    grid_name = "fibgrid_12.5"
    cell_fn_format = "{:04d}.nc"

    @staticmethod
    def fn_read_fmt(timestamp, sat="[ABC]"):
        sat = sat.upper()
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": sat,
            "placeholder": "*",
            "placeholder1": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp):
        return {}


class AscatH130Swath(AscatSwathProduct):
    fn_pattern = "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP{sat}-6.25km-H130_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc"
    sf_pattern = {}
    date_field_fmt = "%Y%m%d%H%M%S"
    grid_name = "fibgrid_6.25"
    cell_fn_format = "{:04d}.nc"

    @staticmethod
    def fn_read_fmt(timestamp, sat="[ABC]"):
        sat = sat.upper()
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": sat,
            "placeholder": "*",
            "placeholder1": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp):
        return {}


class AscatH139Swath(AscatSwathProduct):
    fn_pattern = "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP{sat}-12.5km-H139_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc"
    sf_pattern = {}
    date_field_fmt = "%Y%m%d%H%M%S"
    grid_name = "fibgrid_12.5"
    cell_fn_format = "{:04d}.nc"

    @staticmethod
    def fn_read_fmt(timestamp, sat="[ABC]"):
        sat = sat.upper()
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": sat,
            "placeholder": "*",
            "placeholder1": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp):
        return {}


class AscatSIG0Swath6250m(AscatSwathProduct):
    """
    Class for reading ASCAT sigma0 swath data and writing it to cells.
    """
    fn_pattern = "W_IT-HSAF-ROME,SAT,SIG0-ASCAT-METOP{sat}-6.25_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc"
    sf_pattern = {"satellite_folder": "metop_[abc]", "year_folder": "{year}"}
    date_field_fmt = "%Y%m%d%H%M%S"
    grid_name = "fibgrid_6.25"
    cell_fn_format = "{:04d}.nc"

    @staticmethod
    def fn_read_fmt(timestamp, sat="[ABC]"):
        """
        Format a timestamp to search as YYYYMMDD*, for use in a regex
        that will match all files covering a single given date.

        Parameters
        ----------
        timestamp: datetime.datetime
            Timestamp to format

        Returns
        -------
        dict
            Dictionary of formatted strings
        """
        sat = sat.upper()
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": sat,
            "placeholder": "*",
            "placeholder1": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp, sat="[abc]"):
        sat = sat.lower()
        return {
            "satellite_folder": {
                "satellite": f"metop_{sat}"
            },
            "year_folder": {
                "year": f"{timestamp.year}"
            },
        }


class AscatSIG0Swath12500m(AscatSwathProduct):
    """
    Class for reading and writing ASCAT sigma0 swath data.
    """
    fn_pattern = "W_IT-HSAF-ROME,SAT,SIG0-ASCAT-METOP{sat}-12.5_C_LIIB_{placeholder}_{placeholder1}_{date}____.nc"
    sf_pattern = {"satellite_folder": "metop_[abc]", "year_folder": "{year}"}
    date_field_fmt = "%Y%m%d%H%M%S"
    grid_name = "fibgrid_12.5"
    cell_fn_format = "{:04d}.nc"

    @staticmethod
    def fn_read_fmt(timestamp, sat="[ABC]"):
        """
        Format a timestamp to search as YYYYMMDD*, for use in a regex
        that will match all files covering a single given date.

        Parameters
        ----------
        timestamp: datetime.datetime
            Timestamp to format

        Returns
        -------
        dict
            Dictionary of formatted strings
        """
        sat = sat.upper()
        return {
            "date": timestamp.strftime("%Y%m%d*"),
            "sat": sat,
            "placeholder": "*",
            "placeholder1": "*"
        }

    @staticmethod
    def sf_read_fmt(timestamp, sat="[abc]"):
        sat = sat.lower()
        return {
            "satellite_folder": {
                "satellite": f"metop_{sat}"
            },
            "year_folder": {
                "year": f"{timestamp.year}"
            },
        }


cell_io_catalog = {
    "H129": AscatH129Cell,
    "H130": AscatH130Cell,
    "H122": AscatH122Cell,
    "H121": AscatH121Cell,
    "H139": AscatH139Cell,
    "H29": AscatH29Cell,
    "SIG0_6.25": AscatSIG0Cell6250m,
    "SIG0_12.5": AscatSIG0Cell12500m,
    "ERSH": ErsHCell,
    "ERSN": ErsNCell,
}

swath_io_catalog = {
    "H129": AscatH129Swath,
    "H130": AscatH130Swath,
    "H122": AscatH122Swath,
    "H121": AscatH121Swath,
    "H139": AscatH139Swath,
    "H29": AscatH29Swath,
    "SIG0_6.25": AscatSIG0Swath6250m,
    "SIG0_12.5": AscatSIG0Swath12500m,
}

swath_fname_regex_lookup = {
    "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP[ABC]-6.25km-H129_C_LIIB_.*_.*_.*____.nc":
        "H129",
    "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP[ABC]-6.25km-H130_C_LIIB_.*_.*_.*____.nc":
        "H130",
    "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP[ABC]-6.25km-H122_C_LIIB_.*_.*_.*____.nc":
        "H122",
    "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP[ABC]-12.5km-H121_C_LIIB_.*_.*_.*____.nc":
        "H121",
    "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP[ABC]-12.5km-H139_C_LIIB_.*_.*_.*____.nc":
        "H139",
    "W_IT-HSAF-ROME,SAT,SSM-ASCAT-METOP[ABC]-12.5km-H29_C_LIIB_.*_.*_.*____.nc":
        "H29",
    "W_IT-HSAF-ROME,SAT,SIG0-ASCAT-METOP[ABC]-6.25_C_LIIB_.*_.*_.*____.nc":
        "SIG0_6.25",
    "W_IT-HSAF-ROME,SAT,SIG0-ASCAT-METOP[ABC]-12.5_C_LIIB_.*_.*_.*____.nc":
        "SIG0_12.5",
}


def get_swath_product_id(filename):
    for pattern, swath_product_id in swath_fname_regex_lookup.items():
        if re.match(pattern, filename):
            return swath_product_id
    raise ValueError(
        f"Filename {filename} does not match any known swath product ID pattern."
    )
